CON
  QTR_EMITTERS_OFF = 0
  QTR_EMITTERS_ON  = 1
'  QTR_EMITTERS_ON_AND_OFF = 2  not implemented yet  read on in sensorvalues read off in offvalues sensorValues += maxvalue - offValues
  cntMin     = 400      ' Minimum waitcnt value to prevent lock-up
  timeout = 2000                'uS
{{ timeout' specifies the length of time in uS beyond
   which you consider the sensor reading completely black.  That is to say,
   if the pulse length for a pin exceeds 'timeout', pulse timing will stop
   and the reading for that pin will be considered full black.
   It is recommended that you set timeout to be between 1000 and
   3000 us, depending on things like the height of your sensors and
   ambient lighting.  Using timeout allows you to shorten the
   duration of a sensor-reading cycle while still maintaining
   useful analog measurements of reflectance. }}
   WHITE_LINE = TRUE
   BLACK_LINE = FALSE
OBJ
  Rom : "Propeller Eeprom"
VAR
  long lastLineValue, delay  
  long sensorValue[8]
  '' ***** do not use any word varibles.  Next two lines must stay together for eeprom write/read
  long calibratedMinOn[8], calibratedMaxOn[8], calibratedMinOff[8], calibratedMaxOff[8] 
  byte calibrated[2]
  byte ledPin, sensorPin, numberSensors, lastSensor, uS, sensorState 

PUB Start(_ledPin, _sensorPin, _numberSensors)  
  ledPin := _ledPin              ' ********** what to do for 1RC which has no led pin????    -1 ??
  sensorPin := _sensorPin
  numberSensors := _numberSensors
  lastSensor := sensorPin + numberSensors - 1
  numberSensors := _numberSensors
  if ledPin => 0 
    outa[ledPin]~~                'output
  uS := clkfreq / 1_000_000                  ' Clock cycles for 1 us 
  delay :=  TIMEOUT * uS
  bytefill(@calibrated, 0, 2)                           ' initialize to FALSE
  lastLineValue := 0               'assume initially that the line is left.

PUB GetCal(_sensor)
  return calibrated[_sensor]
    
PUB ReadRaw(readMode)   | now, value, lastValue, i, changeTime, oneValue
'' Reads the sensors and places raw input into sensoValue array
'' Parameter readMode controls the LED values are  QTR_EMITTERS_OFF or QTR_EMITTERS_ON
  bytefill(@sensorValue, 0, 32)                        ' set sensor values array to 0
  if ledPin => 0
    if (readMode ==  QTR_EMITTERS_ON) 'or (readMode == QTR_EMITTERS_ON_AND_OFF)
      outa[ledPin]~~
    else
      outa[ledPin]~
     
  dira[sensorPin..lastSensor]~~                         ' output
  outa[sensorPin..lastSensor]~~                         ' high
  Pause_uS(10)
  dira[sensorPin..lastSensor]~                          ' input
  lastValue := ina[sensorPin..lastSensor]               ' last state of sensor pins should be all 1's here
  now :=  cnt
  repeat until (cnt - now > delay)
    if (value == lastValue)
      next
    else
      changeTime := cnt
      repeat i from 0 to numberSensors - 1
        if (sensorValue[i] == 0) and (ina[sensorPin + i] == 0)
          sensorValue[i] := changeTime - now
  repeat i from 0 to numberSensors - 1
    if sensorValue[i] == 0
      sensorValue[i] := delay                           ' sensorValues are in clkticks

PUB ReadCalibrated(readMode)  | i, x, denominator
'' Reads the sensors and places a calibrated value (0-1000) into sensorValue returns TRUE
'' If not calibrated, returns FALSE and fills sensorValue array with -1
  if readMode == QTR_EMITTERS_ON
    if calibrated[readMode] == False
      longfill(@sensorValue,-1,8)                         ' if not calibrated set sensor value to impossible value
     return  FALSE
    ReadRaw(readMode)                              ' get raw values then calibrate
    repeat i from 0 to numberSensors - 1
      x := 0
      denominator := calibratedMaxOn[i] - calibratedMinOn[i]
      if denominator <> 0
        x := (sensorValue[i] - calibratedMinOn[i]) * 1000 /denominator  ' **** need floating point??
      sensorValue[i] := 0 #> x <# 1000                               ' limit value to 0 -- 1000
    return TRUE
  elseif readmode == QTR_EMITTERS_OFF
    if calibrated[readMode] == False
      longfill(@sensorValue,-1,8)                         ' if not calibrated set sensor value to impossible value
      return  FALSE
    ReadRaw(readMode)                              ' get raw values then calibrate
    repeat i from 0 to numberSensors - 1
      x := 0
      denominator := calibratedMaxOff[i] - calibratedMinOff[i]
      if denominator <> 0
        x := (sensorValue[i] - calibratedMinOff[i]) * 1000 /denominator  ' **** need floating point??
      sensorValue[i] := 0 #> x <# 1000                               ' limit value to 0 -- 1000
    return TRUE
PUB LinePos(readMode, whiteLine)  | onLine, average, sum, i, value
{{Operates the same as read calibrated, but also returns an
  estimated position of the robot with respect to a line. The
  estimate is made using a weighted average of the sensor indices
  multiplied by 1000, so that a return value of 0 indicates that
  the line is directly below sensor 0, a return value of 1000
  indicates that the line is directly below sensor 1, 2000
  indicates that it's below sensor 2000, etc.  Intermediate
  values indicate that the line is between two sensors.  The
  formula is:
 
     0*value0 + 1000*value1 + 2000*value2 + ...
    --------------------------------------------
          value0  +  value1  +  value2 + ...

  By default, this function assumes a dark line (high values)
  surrounded by white (low values).  If your line is light on
  black, set the optional second argument white_line to true.  In
  this case, each sensor value will be replaced by (1000-value)
  before the averaging.
  If sensors are not calibrate LinePos returns -1}}
  onLine := FALSE
  average := 0
  sum := 0
  sensorState := 0
  if ReadCalibrated(readMode)
    repeat i from 0 to numberSensors -1
      value := sensorValue[i]
      if  whiteLine
        value := 1000 - value
      if value > 200       ' keep track of whether we see the line at all
        onLine := TRUE
      if value > 100   '50        ' only average in values that are above a noise threshold
        average += value * (i * 1000)
        sum += value
        sensorState |= |< i           ' set bit to reflect sensor sees line
    if onLine == FALSE
      if lastLineValue < (numberSensors - 1) * 500
        return 0
      else
        return (numberSensors - 1) * 1000
    lastLineValue := average / sum
    return lastLineValue
  else
    lastLineValue := -1
    return lastLineValue   
       
Pub GetSensorState
  return sensorState
       
PUB Calibrate(readMode, calibrationTime)   | i, now
{{  During the calibration time move the sensor array so each individual
    sensor will see both lightest light and darkest dark.
    Calibrate must be called both on and off if both reads are to be used.  }}
  calibrationTime <#= 26                              ' limit to prevent rollover 
  calibrationTime *= clkfreq                            ' calibration time in ticks
  if readMode == QTR_EMITTERS_OFF
    repeat i from 0 to numberSensors - 1
      calibratedMinOff[i] := delay                   ' Initialize the max and min calibrated values to values that                                                 
      calibratedMaxOff[i] :=  0                      ' will cause the first reading to update them.
    now := cnt
    repeat until (cnt - now > calibrationTime)
      ReadRaw(readMode)
        repeat i from 0 to numberSensors - 1
          if sensorValue[i] > calibratedMaxOff[i]
            calibratedMaxOff[i] := sensorValue[i]
          if sensorValue[i] < calibratedMinOff[i]
            calibratedMinOff[i] := sensorValue[i]
    calibrated[readMode] := TRUE         
  if readMode == QTR_EMITTERS_ON
    repeat i from 0 to numberSensors - 1
      calibratedMinOn[i] := delay                   ' Initialize the max and min calibrated values to values that                                                 
      calibratedMaxOn[i] :=  0                      ' will cause the first reading to update them.  
    now := cnt
    repeat until (cnt - now > calibrationTime)
      ReadRaw(readMode)
        repeat i from 0 to numberSensors - 1
          if sensorValue[i] > calibratedMaxOn[i]
            calibratedMaxOn[i] := sensorValue[i]
          if sensorValue[i] < calibratedMinOn[i]
            calibratedMinOn[i] := sensorValue[i]
    calibrated[readMode] := TRUE


PUB SaveCalibration
  Rom.FromRam(@calibratedMinOn[0], @calibrated[1], $8000)

PUB ReadCalibration : status
'' returns 0 if no calibration 1 if LED off calibrated, 2 if LED on calibrated, 3 if both calibrated
  Rom.ToRam(@calibratedMinOn[0], @calibrated[1], $8000)
  if calibrated[0] 
    status := 1
'  if calibrated[1] == TRUE
  if calibrated[1]
    status += 2
PUB GetValue(_sensor)
'' Returns the last value read by sensor.  Normal usage is to call one of the read methods (ReadRaw,
'' ReadCalibrated or LinePos) then call GetValue
  return sensorValue[_sensor]

PUB Stop(_threshold, _whiteline) | i, value
'' if all sensors see the line return TRUE
  result := TRUE
  repeat i from 0 to numberSensors - 1
    if _whiteline
      value := 1000 - sensorValue[i]
    else
      value := sensorValue[i]
    if value < _threshold
      result := FALSE
      return
              
PUB LedOn
'' Not needed for normal use.  Can be used for testing LED
  if ledPin => 0
    outa[ledPin]~~

PUB LedOff
'' Not needed for normal use.  Can be used for testing LED
  if ledPin => 0
    outa[ledPin]~
  
PRI PAUSE_uS(Duration) | clkCycles       ' from BS2_Functions
{{
   Causes a pause for the duration in uS
   Smallest value is 20 at clkfreq = 80Mhz
   Largest value is around 50 seconds at 80Mhz.
     BS2.Pause_uS(1000)   ' 1 mS pause
}}
   clkCycles := Duration * uS #> cntMin                    ' duration * clk cycles for us
                                                           ' - inst. time, min cntMin 
   waitcnt(clkcycles + cnt)                                ' wait until clk gets there                             